@@ -30,13 +30,16 @@ __version__ = "$Id$"
 
 import logging
 import gobject
+import sha
+from base64 import b64encode
 
 import protocol
 import transport
 import structure
+import storage
 import msnp2p
 import consts
-from consts import PresenceStatus
+from consts import PresenceStatus, ServerType
 
 logger = logging.getLogger('client')
 
@@ -80,8 +83,13 @@ class UserProfile(gobject.GObject):
                 "Presence",
                 "The presence to show to others",
                 "",
-                gobject.PARAM_READWRITE)
+                gobject.PARAM_READWRITE),
 
+            "display-picture":  (gobject.TYPE_STRING,
+                "Display picture",
+                "The display picture to show to others",
+                "",
+                gobject.PARAM_READABLE),
             }
 
     def __init__(self, client, account):
@@ -94,65 +102,102 @@ class UserProfile(gobject.GObject):
             @param account: the user account
             @type account: (passport: string, password:string)"""
         gobject.GObject.__init__(self)
-        self.__client = client
-        self.__passport = account[0] 
-        self.__password = account[1]
-        self.__friendly_name = account[0]
-        self.__personal_message = ""
-        self.__presence = PresenceStatus.OFFLINE # current presence
-        self.__profile = ""
+        self._client = client
+        self._passport = account[0] 
+        self._password = account[1]
+        self._friendly_name = account[0]
+        self._personal_message = ""
+        self._presence = PresenceStatus.OFFLINE # current presence
+        self._profile = ""
+        self._display_picture = ""
+        self._msnobject = ""
 
     def __get_profile(self):
-        return self.__profile
+        return self._profile
     profile = property(__get_profile,
         doc="The server profile for this account")
     
     def __set_friendly_name(self, friendly_name):
-        self.__client._protocol.set_friendly_name(friendly_name)
+        self._client._protocol.set_friendly_name(friendly_name)
     def __get_friendly_name(self):
-        return self.__friendly_name
+        return self._friendly_name
     friendly_name = property(__get_friendly_name, __set_friendly_name,
         doc="Friendly name to be shown to your contacts")
         
     def __set_presence(self, presence):
-        self.__client._protocol.set_presence(presence)
+        self._client._protocol.set_presence(presence)
     def __get_presence(self):
-        return self.__presence
+        return self._presence
     presence = property(__get_presence, __set_presence,
         doc="User presence")
 
     def __set_personal_message(self, personal_message):
-        self.__client._protocol.set_personal_message(personal_message)
+        self._client._protocol.set_personal_message(personal_message)
     def __get_personal_message(self):
-        return self.__personal_message
+        return self._personal_message
     personal_message = property(__get_personal_message, __set_personal_message,
         doc="User personal message")
-        
+
+    def __set_display_picture(self, display_picture):
+        if display_picture == self._display_picture:
+            return
+        if display_picture:
+            sha1 = sha.new(display_picture)
+            creator = self._passport
+            size = len(display_picture)
+            type = 3
+            location = sha1.hexdigest() + ".tmp"
+            friendly = "AAA="
+            sha1d = b64encode(sha1.digest())
+            fields = [creator, size, type, location, friendly, sha1d]
+            sha1c = "Creator%sSize%uType%uLocation%sFriendly%sSHA1D%s " % tuple(fields)
+            sha1c = b64encode(sha.new(sha1c).digest())
+            fields.append(sha1c)
+            xml = ('<msnobj Creator="%s" Size="%d" Type="%d" ' + 
+                'Location="%s" Friendly="%s" SHA1D="%s" SHA1C="%s"/>') % tuple(fields)
+            self._msnobject = xml
+            self._display_picture = display_picture
+        else:
+            self._display_picture = ""
+            self._msnobject = ""
+        self._client._protocol.set_presence(self._presence, self._msnobject)
+    def __get_display_picture(self):
+        return self._display_picture
+    display_picture = property(__get_display_picture, __set_display_picture,
+        doc="User display picture")
+
+    def __get_msnobject(self):
+        return self._msnobject
+    msnobject = property(__get_msnobject,
+        doc="User msnobject")
+
     def do_get_property(self, pspec):
         if pspec.name == "passport":
-            return self.__passport
+            return self._passport
         elif pspec.name == "password":
-            return self.__password
+            return self._password
         elif pspec.name == "friendly-name":
-            return self.__friendly_name
+            return self._friendly_name
         elif pspec.name == "personal-message":
-            return self.__personal_message
+            return self._personal_message
         elif pspec.name == "presence":
-            return self.__presence
+            return self._presence
         elif pspec.name == "profile":
-            return self.__profile
+            return self._profile
+        elif pspec.name == "display-picture":
+            return self._display_picture
         else:
             raise AttributeError, "unknown property %s" % pspec.name
 
     def do_set_property(self, pspec, value):
         if pspec.name == "friendly-name":
-            self.__friendly_name = value
+            self._friendly_name = value
         elif pspec.name == "personal-message":
-            self.__personal_message = value
+            self._personal_message = value
         elif pspec.name == "presence":
-            self.__presence = value
+            self._presence = value
         elif pspec.name == "profile":
-            self.__profile = value
+            self._profile = value
         else:
             raise AttributeError, "unknown property %s" % pspec.name
 
@@ -169,7 +214,9 @@ class Client(object):
     @group Callbacks: on_*"""
 
     def __init__(self, server, account, proxies={},
-            initial_status=consts.PresenceStatus.ONLINE):
+            initial_status=consts.PresenceStatus.ONLINE,
+            transport_class=transport.DirectConnection,
+            storage_class=storage.MemoryStorage):
         """Initializer
 
             @param server: the Notification server to connect to.
@@ -183,10 +230,19 @@ class Client(object):
 
             @param initial_status: the initial status to login with
             @type initial_status: integer
-            @see L{consts.PresenceStatus}"""        
+            @see L{consts.PresenceStatus}
+            
+            @param transport_class: a L{transport.BaseTransport} derived
+                class, used to make the connection to the MSN service
+                
+            @param storage_class: a storage class, used by pymsn to store
+                and/or cache data.
+            @type storage_class: L{storage.AbstractStorage} derived class"""        
         
         self.profile = UserProfile(self, account)
-        self._transport = transport.DirectConnection(server)
+        self._storage = storage_class
+        self._transport_class = transport_class
+        self._transport = transport_class(server, ServerType.NOTIFICATION, proxies)
         self._proxies = proxies
         
         self._protocol = protocol.NotificationProtocol(self, self._transport,
@@ -286,6 +342,12 @@ class Client(object):
         return self._protocol._get_contacts_by_list(list_type)
     
     ### internal Methods
+    def _register_switchboard(self, switchboard):
+        self._switchboards.append(switchboard)
+
+    def _unregister_switchboard(self, switchboard):
+        self._switchboards.remove(switchboard)
+
     def _request_switchboard(self, callback):
         self._switchboards_callbacks.append(callback)
         self._protocol.request_switchboard()
@@ -467,14 +529,14 @@ class Conversation(msnp2p.P2PTransport):
             assert(key != "")
             assert(session != "")
             self.__invitee = []
-            self._transport = transport.DirectConnection(server)
+            self._transport = self._client._transport_class(server, ServerType.SWITCHBOARD, self._proxies)
             self._switchboard = protocol.SwitchboardProtocol(self._client,
-                    self._transport, [], key, session, client._proxies)
+                    self._transport, [], key, session, self._proxies)
             self.__attach_switchboard()
         else: # invite someone
             assert(invitee)
         self.__requested_sb = False
-
+        self._client._register_switchboard(self)
         gobject.idle_add(self.__connect)
 
     # Public methods
@@ -623,7 +685,7 @@ class Conversation(msnp2p.P2PTransport):
             self.__process_pending_actions()
 
     def __on_switchboard_response(self, server, key):
-        self._transport = transport.DirectConnection(server)
+        self._transport = self._client._transport_class(server, ServerType.SWITCHBOARD, self._proxies)
         self._switchboard = protocol.SwitchboardProtocol(self._client,
                 self._transport, self.__invitee, key, "", self._proxies)
         self.__attach_switchboard()
@@ -661,11 +723,10 @@ class Conversation(msnp2p.P2PTransport):
     def get_p2p_peers(self):
         if self._switchboard != None:
             peers = self._switchboard.users.keys()
-        else:
-            peers = None
-
-        if not peers:
+        elif len(self.__invitee) > 0:
             peers = (self.__invitee[0],)
+        else:
+            peers = []
 
         return [self._client.get_contact_by_passport(peer) for peer in peers]
 
