@@ -30,6 +30,7 @@ polling transport ideal in firewalled environment."""
 __version__ = "$Id$"
 
 import logging
+import base64
 import gobject
 
 import gio
@@ -68,13 +69,22 @@ class BaseTransport(gobject.GObject):
                 (object,)),
             }   
 
-    def __init__(self, server):
+    def __init__(self, server, server_type=ServerType.NOTIFICATION, proxies={}):
         """Connection initialization
         
             @param server: the server to connect to.
-            @type server: (host: string, port: integer)"""
+            @type server: (host: string, port: integer)
+
+            @param server_type: the server that we are connecting to, either
+                Notification or switchboard.
+            @type server_type: L{consts.ServerType}
+
+            @param proxies: proxies that we can use to connect
+            @type proxies: {type: string => L{gio.network.ProxyInfos}}"""
         gobject.GObject.__init__(self)
         self.server = server
+        self.server_type = server_type
+        self.proxies = proxies
         self._transaction_id = 0
     
     def __get_transaction_id(self):
@@ -144,7 +154,11 @@ class BaseTransport(gobject.GObject):
             @param cb_args: callback arguments
             @type cb_args: tuple
         """
-        raise NotImplementedError
+        if transaction_id is not None and transaction_id < 0:
+            transaction_id = self._transaction_id
+        cmd = structure.Command()
+        cmd.build(command, transaction_id, arguments, payload)
+        self.send_command(cmd, increment, callback, cb_args)
 
     def _increment_transaction_id(self):
         """Increments the Transaction ID then return it.
@@ -159,8 +173,8 @@ gobject.type_register(BaseTransport)
 class DirectConnection(BaseTransport):
     """Implements a direct connection to the net without any proxy""" 
 
-    def __init__(self, server):
-        BaseTransport.__init__(self, server)
+    def __init__(self, server, server_type=ServerType.NOTIFICATION, proxies={}):
+        BaseTransport.__init__(self, server, server_type, proxies)
         
         transport = gio.network.TCPClient(server[0], server[1])
         transport.connect("notify::status", self.__on_status_change)
@@ -208,14 +222,6 @@ class DirectConnection(BaseTransport):
         if user_callback:
             user_callback(*user_cb_args)
 
-    def send_command_ex(self, command, arguments=None, payload=None,
-            transaction_id=-1, increment=True, callback=None, cb_args=()):
-        if transaction_id is not None and transaction_id < 0:
-            transaction_id = self._transaction_id
-        cmd = structure.Command()
-        cmd.build(command, transaction_id, arguments, payload)
-        self.send_command(cmd, increment, callback, cb_args)
-
     ### callbacks
     def __on_status_change(self, transport, param):
         status = transport.get_property("status")
@@ -248,176 +254,258 @@ class DirectConnection(BaseTransport):
         self.emit("command-received", cmd)
 
 gobject.type_register(DirectConnection)
- 
-class HTTPPollConnection(BaseTransport):
-    """Implements an http connection to the net without any proxy"""
+
+
+class HTTPPollTransportStatus(object):
+    """Transport status."""
     
-    def __init__(self, server, server_type=ServerType.NOTIFICATION):
-        BaseTransport.__init__(self, server)
+    OPEN = 0
+    """The transport is open."""
+    OPENING_REQUESTED = 1
+    """The transport is being opened due to user request."""
+    OPENING_AUTO = 2
+    """The transport is being opened due to auto reconnection process."""
+    LOST = 3
+    """The transport lost his connection."""
+    CLOSING = 4
+    """The transport is being closed."""
+    CLOSED = 5
+    """The transport is closed."""
 
-        transport = gio.network.TCPClient(server[0], server[1])
-        self.status_id = transport.connect("notify::status", self.__on_status_change)
+class HTTPPollTransactionStatus(object):
+    """Data transfert status"""
+    NONE = 0
+    SENDING = 1
+    RECEIVING = 2
+    IGNORING = 3 # to deal with HTTP/1.1 100 Continue
+
+class HTTPPollConnection(BaseTransport):
+    """Implements an http connection to the net"""
+
+    POLL_DELAY = 3000
+
+    def __init__(self, server, server_type=ServerType.NOTIFICATION, proxies={}):
+        self._target_host = server[0]
+        server = ("gateway.messenger.hotmail.com", 80)
+        BaseTransport.__init__(self, server, server_type, proxies)
+        
+        if 'http' in self.proxies:
+            transport = gio.network.TCPClient(proxies['http'].host, proxies['http'].port)
+        else:
+            transport = gio.network.TCPClient(server[0], server[1])
+        transport.connect("notify::status", self.__on_status_change)
         transport.connect("error", lambda t, msg: self.emit("connection-failure"))
-        # self._die = False
-
-        logger.debug(self.status_id)
-
+    
         receiver = gio.ChunkReceiver(transport)
         receiver.connect("received", self.__on_received)
         
         self._receiver = receiver
         self._receiver.delimiter = "\r\n"
         self._transport = transport
-        self.__resetting = False
 
-        self.command_queue = []
-        self.waiting_for_response = False
-        self.notification = True
-        self.server_type = server_type
+        self._command_queue = []
 
-        # Configuring the polling period
-        gobject.timeout_add(3000, self.__poll)
+        self._transport_status = HTTPPollTransportStatus.CLOSED
+        self._transaction_status = HTTPPollTransactionStatus.NONE
+        self._session_open = False          # used to send Action=open
         
         self.__clear_response_handler()
-    __init__.__doc__ = BaseTransport.__init__.__doc__ + \
-            """
 
-            @param server_type: the server type
-            @type server_type: L{consts.ServerType}"""
+        self._poll_id = None
+        self._disable_poll = True
+        
+    __init__.__doc__ = BaseTransport.__init__.__doc__
 
-    # Connection
+    ### public commands
     def establish_connection(self):
-        logger.debug('<-> Connecting to %s:%d' % self.server )
+        logger.debug('<-> Connecting to %s:%d' % self.server)
         self._transport.open()
 
     def lose_connection(self):
+        self._transport_status = HTTPPollTransportStatus.CLOSING
+        self._session_open = False
+        self._transaction_status = HTTPPollTransactionStatus.NONE
         self._transport.close()
 
-    def reset_connection(self, server=None):
+    def reset_connection(self, server=None): #TODO: HTTPPollTransportStatus.RESETING ?
         if server:
             self._transport.set_property("host", server[0])
             self._transport.set_property("port", server[1])
             self.server = server
-        self.__resetting = True
         self._transport.close()
+        self._session_open = False
+        self._transaction_status = HTTPPollTransactionStatus.NONE
         self._transport.open()
 
-    # Command Sending
-    def send_command(self, command=None, increment=True, callback=None,
-            cb_args=()):
-        if len(self.command_queue) > 0 or self.waiting_for_response:
-            self.command_queue.insert(0, (command, increment, callback, cb_args))
-        else:        
-            self.__send_it(command, increment, callback, cb_args)
-
-    def __send_it(self, command, increment=True, user_callback=None,
-            user_cb_args=()):
-        host = self._transport.get_property("host")
-        strcmd = str(command)
-        lstrcmd = len(strcmd)
-
-        if self.notification:            
-            params = "Action=open&Server=%s&IP=%s" % (self.server_type, host)
-            self.notification = False
-        elif command == None:
-            # Polling the server for queued messages
-            params = "Action=poll&SessionID=%s" % self.session_id
-            strcmd = ""
-            lstrcmd = 0
-        else:
-            params = "SessionID=%s" % self.session_id
-
-        header = ("POST http://%s/gateway/gateway.dll?%s HTTP/1.1\r\n" +\
-                 "Accept: */*\r\n" +\
-                 "Accept-Language: en-us\r\n" +\
-                 "User-Agent: MSMSGS\r\n" +\
-                 "Host: %s\r\n" +\
-                 "Proxy-Connection: Keep-Alive\r\n" +\
-                 "Connection: Keep-Alive\r\n" +\
-                 "Pragma: no-cache\r\n" +\
-                 "Content-Type: application/x-msn-messenger\r\n" +\
-                 "Content-Length: %d\r\n") % (host, params, host, lstrcmd)
-
-        self._transport.send(header + "\r\n" + strcmd)
-        self.waiting_for_response = True
-
-        logger.debug('>>> ' + repr(command))
-        
+    def send_command(self, command, increment=True, callback=None, cb_args=()):
+        self.__queue_command(command, callback, cb_args)
         if increment:
             self._increment_transaction_id()
+
+    def __queue_command(self, command, callback, cb_args):
+        self._command_queue.append((command, callback, cb_args))
+        self.__process_command_queue()
+
+    def __pop_command(self):
+        return self._command_queue.pop(0)
+    
+    def __send_command(self, command, callback=None, cb_args=()):
+        logger.debug('>>> ' + repr(command))
+        host = self.server[0]
+        strcmd = str(command)
         
+        if not self._session_open:
+            params = "Action=open&Server=%s&IP=%s" % (self.server_type, self._target_host)
+            self._session_open = True
+        elif command == None: # Polling the server for queued messages
+            assert(self._transaction_status == HTTPPollTransactionStatus.NONE)
+            params = "Action=poll&SessionID=%s" % self.session_id
+            strcmd = ""
+        else: # new command
+            assert(self._transaction_status == HTTPPollTransactionStatus.NONE)
+            params = "SessionID=%s" % self.session_id
+
+        action = "POST http://%s/gateway/gateway.dll?%s HTTP/1.1" % (host, params)
+        headers = []
+
+        headers.append("Accept: */*")
+        headers.append("Accept-Language: en-us")
+        headers.append("User-Agent: MSMSGS")
+        headers.append("Host: " + host)
+        headers.append("Proxy-Connection: Keep-Alive")
+        headers.append("Connection: Keep-Alive")
+        headers.append("Pragma: no-cache")
+        headers.append("Content-Type: application/x-msn-messenger")
+        headers.append("Content-Length: %d" % len(strcmd))
+        if 'http' in self.proxies and self.proxies['http'].user:
+                auth = base64.encodestring(self.proxies['http'].user + ':' + self.proxies['http'].password)
+                headers.append("Proxy-authorization: Basic " + auth) 
+
+        http_envelope = action + "\r\n" + "\r\n".join(headers) + "\r\n"
+
+        self._transaction_status = HTTPPollTransactionStatus.SENDING
+        command_sent_cb_args = (command, callback, cb_args)
+        self._transport.send(http_envelope + "\r\n" + strcmd,
+                self.__on_command_sent, command_sent_cb_args)
+        blob = http_envelope + "\r\n" + strcmd
+        
+        #for line in blob.split('\r\n'):
+        #    print '\t>>> ', line
+        #print '----------------------------------------------------------'
+
+    def __process_command_queue(self):
+        s = HTTPPollTransportStatus
+        if self._transport_status == s.OPEN:
+            if self._transaction_status != HTTPPollTransactionStatus.NONE:
+                return
+
+            if len(self._command_queue) > 0:
+                cmd = self.__pop_command()
+                self.__send_command(*cmd)
+            elif not self._disable_poll:
+                self.__send_command(None) # Poll (WARN: don't use self.__poll())
+        elif self._transport_status == s.CLOSED or \
+                self._transport_status == s.LOST:
+            self.establish_connection()
+
+    def __poll(self):
+        if len(self._command_queue) == 0:
+            self.__process_command_queue()
+        return True
+
+    ### callbacks
+    def __on_command_sent(self, command, user_callback, user_cb_args):
+        self.emit("command-sent", command)
         if user_callback:
             user_callback(*user_cb_args)
 
-    def send_command_ex(self, command, arguments=None, payload=None,
-            transaction_id=-1, increment=True, callback=None, cb_args=()):
-        if transaction_id is not None and transaction_id < 0:
-            transaction_id = self._transaction_id
-        cmd = structure.Command()
-        cmd.build(command, transaction_id, arguments, payload)
-        self.send_command(cmd, increment, callback, cb_args)
-
-    ### callbacks
     def __on_status_change(self, transport, param):
         status = transport.get_property("status")
+        s = HTTPPollTransportStatus
         if status == gio.STATUS_OPEN:
-            if self.__resetting:
-                self.emit("connection-reset")
-                self.__resetting = False
-            self.emit("connection-success")
-        elif status == gio.STATUS_CLOSED:
-            if not self.__resetting:
-                self.emit("connection-lost")
+            if self._transport_status == s.OPENING_REQUESTED:
+                self._transport_status = s.OPEN
+                self.emit("connection-success")
+                self._poll_id = gobject.timeout_add(self.POLL_DELAY, self.__poll)
+            elif self._transport_status == s.OPENING_AUTO:
+                self._transport_status = s.OPEN
+                self.__process_command_queue()
 
-    def __poll(self):
-        self.send_command()
-        return True
+        elif status == gio.STATUS_OPENING:
+            if self._transport_status == s.CLOSED:
+                self._transport_status = s.OPENING_REQUESTED
+            elif self._transport_status == s.LOST:
+                self._transport_status = s.OPENING_AUTO
+
+        elif status == gio.STATUS_CLOSED:
+            if self._transport_status == s.OPEN:
+                self._transport_status = s.LOST
+            elif self._transport_status == s.CLOSING:
+                self._transport_status = s.CLOSED
+                self.emit("connection-lost")
+                self._disable_poll = True
+                gobject.source_remove(self._poll_id)
 
     def __on_received(self, receiver, chunk):
-        if chunk[:4] == 'HTTP':
-            # If required test connection status here. See :
-            # http://www.w3.org/Protocols/rfc2616/rfc2616-sec8.html
-            # http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.1.1
-            pass
-        elif self.waiting_for_data:
+        #print '\t<<< ', chunk
+        if self._transaction_status == HTTPPollTransactionStatus.SENDING and \
+                chunk[:4] == 'HTTP': # HTTP status line
+            chunk = chunk.split()
+            if chunk[1] == '100':
+                self._transaction_status = HTTPPollTransactionStatus.IGNORING
+            elif chunk[1] == '200':
+                self._transaction_status = HTTPPollTransactionStatus.RECEIVING
+            else:
+                self.emit("connection-failure")
+                self.lose_connection()
+        elif not self.http_body_pending: # headers
+            if len(chunk) != 0:
+                header, value =  [p.strip() for p in chunk.split(':', 1)]
+                if header == 'Content-Length':
+                    self.content_length = int(value)
+                elif header == 'X-MSN-Messenger':
+                    for elem in value.split(';'):
+                        key, val =  [p.strip() for p in elem.split('=', 1)]
+                        if key == 'SessionID':
+                            self.session_id = val
+                        elif key == 'GW-IP':
+                            self.server = (val, self.server[1])
+                            #if 'http' not in self.proxies: # TODO: shall we reset the connection ?
+                            #    self._transport.set_property("host", val)
+                        elif key == 'Session'and val == 'close':
+                            self.emit("connection-lost")
+                            self.lose_connection()
+            else: # empty line separating headers from body
+                if self._transaction_status == HTTPPollTransactionStatus.IGNORING:
+                    self._transaction_status = HTTPPollTransactionStatus.SENDING
+                else:
+                    if self.content_length > 0:
+                        self.http_body_pending = True
+                        self._receiver.delimiter = self.content_length
+                    else:
+                        self.__clear_response_handler()
+        else: # HTTP body
             self.data = chunk
-        elif len(chunk) == 0:
-            self._receiver.delimiter = self.content_length
-            self.waiting_for_data = True
-        else:
-            header, content =  [p.strip() for p in chunk.split(':', 1)]
-            if header == 'Content-Length':
-                self.content_length = int(content)
-            elif header == 'X-MSN-Messenger':
-                for elem in content.split(';'):
-                    key, value =  [p.strip() for p in elem.split('=', 1)]
-                    if key == 'SessionID':
-                        self.session_id = value
-                    elif key == 'GW-IP':
-                        self._transport.set_property("host", value)
-                    elif key == 'Session'and value == 'close':
-                        # Connection is lost before receiving the close message
-                        # TODO : check why in gio and then debug
-                        # self._die = True
-                        self.lose_connection()
-
-        if self.waiting_for_data:
             if self.content_length == 0:
-                logger.debug('>>> Poll : no message queued on server')
+                # The message was an empty response to a poll,
+                # there is nothing to retrieve from the server
+                pass
             elif len(self.data) != 0:
                 while len(self.data) != 0:
                     self.data = self.__extract_command(self.data)
-            self.waiting_for_response = False
+                    
+            # We got a complete response from the server, we can
+            # send the next command or reconnect to the server
+            # if necessary
             self.__clear_response_handler()
-            self._receiver.delimiter = '\r\n'
-            if len(self.command_queue) > 0:
-                (cmd, inc, callback, cb_args) = self.command_queue.pop()
-                self.__send_it(cmd, inc)
+            self.__process_command_queue()
 
     def __extract_command(self, data):
         first, rest = data.split('\r\n', 1)
         cmd = structure.Command()
         cmd.parse(first.strip())
+        if cmd.name == 'USR' and cmd.arguments[0] == 'OK':
+            self._disable_poll = False
         if cmd.name in structure.Command.PAYLOAD_COMMANDS:
             payload_len = int(cmd.arguments[-1])
             if payload_len > 0:
@@ -431,8 +519,10 @@ class HTTPPollConnection(BaseTransport):
             return rest
 
     def __clear_response_handler(self):
-        self.waiting_for_data = False
+        self.http_body_pending = False
         self.data = ''
         self.content_length = 0
-            
+        self._receiver.delimiter = '\r\n'
+        self._transaction_status = HTTPPollTransactionStatus.NONE
+
 gobject.type_register(HTTPPollConnection)
